/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package dpop;

import java.util.*;
import java.lang.*;
import java.io.*;

/**
 *
 * @author Andrea
 */
public class BinaryConstraint implements Constraint {
    private Variable _n1;
    private Variable _n2;
    private TaggedDictionary<Couple, TaggedDictionary<Couple, Value>> _constraint;

    private Value _default_value;

    BinaryConstraint(Variable n1, Variable n2) {
        _n1 = n1;
        _n2 = n2;
        
        _constraint = new TaggedDictionary<Couple, TaggedDictionary<Couple, Value>>();
        _default_value = new IntegerValue(new Integer(0));
    }

    //creates a new constraint with new nodes and same couples
    BinaryConstraint(BinaryConstraint bc, Variable new1, Variable new2) {
        _n1 = new1;
        _n2 = new2;

        _constraint = new TaggedDictionary<Couple, TaggedDictionary<Couple, Value>>();
        _default_value = new IntegerValue(new Integer(0));

        Iterator<Couple> it1 = bc._constraint.keySet().iterator();
        while (it1.hasNext()) {
            Couple current = it1.next();
            Couple newcouple = new Couple(new1, current.getValue());

            TaggedDictionary<Couple, Value> d = new TaggedDictionary<Couple, Value>();
            TaggedDictionary<Couple, Value> old_d = bc._constraint.get(current);
            Iterator<Couple> it2 = old_d.keySet().iterator();

            while (it2.hasNext()) {
                Couple current2 = it2.next();
                Couple newcouple2 = new Couple(new2, current2.getValue());

                d.put(newcouple2, (Value)old_d.get(current2));
            }

            _constraint.put(newcouple, d);
        }
    }

    public void addConstraint(Couple node1_value, Couple node2_value, Value constraint_value) {
        TaggedDictionary<Couple, Value> d;

        if (!_constraint.containsKey(node1_value)) {
            d = new TaggedDictionary<Couple, Value>();
            _constraint.put(node1_value, d);
        }
        else {
            d = (TaggedDictionary<Couple, Value>)_constraint.get(node1_value);
        }

        //the constraint (node1_value, node2_value) -> constraint_value
        //can be set only if there isn't already a constraint value
        if (!d.containsKey(node2_value)) {
            d.put(node2_value, constraint_value);
        }
    }

    public Value getConstraint(Couple[] couples) {
        Couple node1_value;
        Couple node2_value;

        if (couples.length == 2) {
            node1_value = couples[0];
            node2_value = couples[1];
        }
        else {
            return null;
        }

        if (_constraint.containsKey(node1_value)) {
            TaggedDictionary<Couple, Value> d = (TaggedDictionary<Couple, Value>) _constraint.get(node1_value);

            if (d.containsKey(node2_value)) {
                return (Value)d.get(node2_value);
            }
            else {
                //if the couple is not present, return the default value
                //for the weight
                return _default_value;
            }
        }
        else if (_constraint.containsKey(node2_value)) {
            TaggedDictionary<Couple, Value> d = (TaggedDictionary<Couple, Value>) _constraint.get(node2_value);

            if (d.containsKey(node1_value)) {
                return (Value)d.get(node1_value);
            }
            else {
                //if the couple is not present, return the default value
                //for the weight
                return _default_value;
            }
        }
        else {
            //if the couple is not present, return the default value
            //for the weight
            return _default_value;
        }
    }

    public ArrayList<Variable> getVariables() {
        ArrayList<Variable> res = new ArrayList<Variable>();
        res.add(_n1);
        res.add(_n2);
        
        return res;
    }

    @Override
    public String toString() {
        String res = "";
        Iterator<Couple> k1 = _constraint.keySet().iterator();

        while (k1.hasNext()) {
            Couple current1 = k1.next();
            TaggedDictionary<Couple, Value> sec = _constraint.get(current1);

            Iterator<Couple> k2 = sec.keySet().iterator();

            while (k2.hasNext()) {
                Couple current2 = k2.next();
                Value weight = sec.get(current2);

                res = res + "[(" + current1.getValue().toString() + ", " + current2.getValue().toString() + ") -> " + weight.toString() + "] ";
            }
        }

        return res;
    }
}
