/*
 * DPOPApp.java
 */

package dpop;

import org.jdesktop.application.Action;
import org.jdesktop.application.Application;
import org.jdesktop.application.SingleFrameApplication;

import java.util.*;
import java.lang.*;
import java.io.*;
import javax.swing.*;

/**
 * The main class of the application.
 */
public class DPOPApp extends SingleFrameApplication {
    private Interface _view;
    private Handler _handler;

    /**
     * At startup create and show the main frame of the application.
     */
    @Override protected void startup() {
        _view = new Interface();
        _handler = new Handler(_view);

        _view.setLocationRelativeTo(null);
        
        show(_view);
    }

    /**
     * This method is to initialize the specified window by injecting resources.
     * Windows shown in our application come fully initialized from the GUI
     * builder, so this additional configuration is not needed.
     */
    @Override protected void configureWindow(java.awt.Window root) {
    }

    /**
     * A convenient static getter for the application instance.
     * @return the instance of DPOPApp
     */
    public static DPOPApp getApplication() {
        return Application.getInstance(DPOPApp.class);
    }

    /**
     * Main method launching the application.
     */
    public static void main(String[] args) {
        try {
            launch(DPOPApp.class, args);
        }
        catch (Exception e) {
            javax.swing.JOptionPane.showMessageDialog(null, "The application has generated the following exception:\n" + e.getMessage() + "\n\nThe application will now close", "Error", javax.swing.JOptionPane.ERROR_MESSAGE);
            System.exit(1);
        }
    }

    @Action
    public void start_computation() {
        _handler.start_stop();
    }

    @Action
    public void selectFile() {
        JFileChooser fc = new JFileChooser();
        int returnVal = fc.showOpenDialog(null);//_view.getComponent());

        if (returnVal == JFileChooser.APPROVE_OPTION) {
            File file = fc.getSelectedFile();
            _handler.read_input_file(file.getAbsolutePath(), _view.use_heuristic());
        }
        //_handler.read_input_file("/Users/Andrea/Documents/Uni/Specialistica/Ragionamento Automatico/Progetto/adopt/solver/test.txt", _view.use_heuristic());
    }

    @Action
    public void view_source_graph() {
        ViewGraph vg = new ViewGraph();
        vg.readGraph(_handler.getGraphRoot());
        vg.setVisible(true);
    }

    @Action
    public void view_pseudotree() {
        ViewGraph vg = new ViewGraph();
        vg.readPseudoTree(_handler.getPseudoTreeRoot());
        vg.setVisible(true);
    }
}
