/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package dpop;

import java.util.*;
import java.lang.*;
import java.io.*;

/**
 *
 * @author Andrea
 */
public class GraphNode extends Variable {
    private Handler _handler;

    private TaggedDictionary<GraphNode, BinaryConstraint> _neighbors;
    private boolean _visited;
    private PseudoTreeNode _associated_pseudotree_node;

    GraphNode(Object id, ArrayList<Value> domain, Handler h) {
        super(id, domain);

        _handler = h;
        _visited = false;
        _neighbors = null;
    }

    public void add_neighbor(GraphNode n, BinaryConstraint bc) {
        if (_neighbors == null) {
            _neighbors = new TaggedDictionary<GraphNode, BinaryConstraint>(this);
        }

        if (!_neighbors.containsKey(n)) {
            _neighbors.put(n, bc);
            n.add_neighbor(this, bc);
        }
    }

    public int getNeighborsCount() {
        return (_neighbors == null) ? 0 : _neighbors.size();
    }

    public Iterator<GraphNode> getNeighbors() {
        return (_neighbors == null) ? new ArrayList<GraphNode>().iterator() : _neighbors.keySet().iterator();
    }

    public boolean isIsolated() {
        return (_neighbors == null);
    }

    /*
     * Makes a dfs visit on the graph and returns the associated pseudo-tree
     * using the heuristic "most connected node" to select the starting point
     */
    public PseudoTreeNode dfs_visit_mcn() {
        ArrayList<GraphNode> parents = new ArrayList<GraphNode>();
        return _dfs_visit_mcn(parents);
    }

    //do a dfs visit to create a pseudo tree and returns the subtree rooted
    //at the node returned. Uses the most connected node heuristic
    private PseudoTreeNode _dfs_visit_mcn(ArrayList<GraphNode> parents) {
        _visited = true;
        
        Iterator<GraphNode> it_neighbors = getNeighbors();

        //orders first its neighbors with the "most connected node first" criteria
        ArrayList<GraphNode> ordered_neighbors = new ArrayList<GraphNode>();
        while (it_neighbors.hasNext()) {
            GraphNode current = it_neighbors.next();
            //find the place of the current node
            int i;
            for (i = 0; i < ordered_neighbors.size(); i++) {
                if (ordered_neighbors.get(i).getNeighborsCount() < current.getNeighborsCount()) {
                    break;
                }
            }

            ordered_neighbors.add(i, current);
        }
        

        PseudoTreeNode current_ptn = new PseudoTreeNode(getId(), getDomain(), _handler);
        it_neighbors = ordered_neighbors.iterator();

        while (it_neighbors.hasNext()) {
            GraphNode n = it_neighbors.next();
            if (!parents.contains(n)) {    //the ancestors must not be visited
                BinaryConstraint bc = _neighbors.get(n);
                
                if (n._visited) {
                    //remake the constraint between GraphNodes with a constraint
                    //between PseudoTreeNodes
                    BinaryConstraint bc_new = new BinaryConstraint(bc, current_ptn, n._associated_pseudotree_node);

                    //it's a pseudo children
                    current_ptn.addPseudo_child(n._associated_pseudotree_node, bc_new);
                }
                else {
                    ArrayList<GraphNode> new_parents = (ArrayList<GraphNode>)parents.clone();
                    new_parents.add(this);

                    PseudoTreeNode child = n._dfs_visit_mcn(new_parents);
                    //remake the constraint between GraphNodes with a constraint
                    //between PseudoTreeNodes
                    BinaryConstraint bc_new = new BinaryConstraint(bc, current_ptn, child);

                    //it's a normal children
                    current_ptn.addChild(child, bc_new);
                }
            }
        }

        //saves the associated pseudo tree node, in this way when a pseudo parent
        //will try to visit this node, since it is already visited, it returns
        //to the pseudo parent its associated pseudo tree node
        _associated_pseudotree_node = current_ptn;

        return current_ptn;
    }

    /*
     * Makes a dfs visit on the graph and returns the associated pseudo-tree
     * without using any heuristic
     */
    public PseudoTreeNode dfs_visit() {
        ArrayList<GraphNode> parents = new ArrayList<GraphNode>();
        return _dfs_visit(parents);
    }

    //do a dfs visit to create a pseudo tree and returns the subtree rooted
    //at the node returned. Uses the most connected node heuristic
    private PseudoTreeNode _dfs_visit(ArrayList<GraphNode> parents) {
        _visited = true;

        Iterator<GraphNode> it_neighbors = getNeighbors();
        PseudoTreeNode current_ptn = new PseudoTreeNode(getId(), getDomain(), _handler);

        while (it_neighbors.hasNext()) {
            GraphNode n = it_neighbors.next();
            if (!parents.contains(n)) {    //the ancestors must not be visited
                BinaryConstraint bc = _neighbors.get(n);

                if (n._visited) {
                    //remake the constraint between GraphNodes with a constraint
                    //between PseudoTreeNodes
                    BinaryConstraint bc_new = new BinaryConstraint(bc, current_ptn, n._associated_pseudotree_node);

                    //it's a pseudo children
                    current_ptn.addPseudo_child(n._associated_pseudotree_node, bc_new);
                }
                else {
                    ArrayList<GraphNode> new_parents = (ArrayList<GraphNode>)parents.clone();
                    new_parents.add(this);

                    PseudoTreeNode child = n._dfs_visit(new_parents);
                    //remake the constraint between GraphNodes with a constraint
                    //between PseudoTreeNodes
                    BinaryConstraint bc_new = new BinaryConstraint(bc, current_ptn, child);

                    //it's a normal children
                    current_ptn.addChild(child, bc_new);
                }
            }
        }

        //saves the associated pseudo tree node, in this way when a pseudo parent
        //will try to visit this node, since it is already visited, it returns
        //to the pseudo parent its associated pseudo tree node
        _associated_pseudotree_node = current_ptn;

        return current_ptn;
    }


    @Override
    public String toString() {
        return "graph node " + getId().toString() + ", domain: " + getDomain().toString();
    }

    public String shortToString() {
        return "x" + getId().toString();
    }
}
