/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package dpop;

import java.util.*;
import java.lang.*;
import java.io.*;

/**
 *
 * @author Andrea
 */
public class PseudoTreeNode extends Variable implements Runnable {
    private Handler _handler;

    private enum Phase {
        StartPropagation,
        UtilPropagation,
        ValuePropagation
    }

    private static int _nthreads = 0;
    private static int _nthreadcalls = 1;
    private int _util_weight;
    private boolean _use_precise_timing;

    //each child node associated with its util sended
    private TaggedDictionary<PseudoTreeNode, TaggedDictionary<Value, Object>> _util_received;
    private TaggedDictionary<Value, Object> _argmax;
    private TaggedDictionary<PseudoTreeNode, Value> _received_solution;
    private Phase _phase;

    private PseudoTreeNode _parent;
    private ArrayList<PseudoTreeNode> _pseudo_parents;
    private TaggedDictionary<PseudoTreeNode, BinaryConstraint> _childs;
    private TaggedDictionary<PseudoTreeNode, BinaryConstraint> _pseudo_childs;

    PseudoTreeNode(Object id, ArrayList<Value> domain, Handler h) {
        super(id, domain);

        _handler = h;

        _clear_stored_values();

        _parent = null;
        _childs = null;
        _pseudo_parents = null;
        _pseudo_childs = null;
    }

    PseudoTreeNode(GraphNode n, Handler h) {
        super(n.getId(), n.getDomain());

        _handler = h;

        _clear_stored_values();

        _parent = null;
        _childs = null;
        _pseudo_parents = null;
        _pseudo_childs = null;
    }

    private void _clear_stored_values() {
        _argmax = new TaggedDictionary<Value, Object>(this);
        _util_received =  new TaggedDictionary<PseudoTreeNode, TaggedDictionary<Value, Object>>(this);
        _received_solution = new TaggedDictionary<PseudoTreeNode, Value>(this);
        _phase = Phase.StartPropagation;
    }

    public Iterator<PseudoTreeNode> get_childs() {
        if (_childs == null) {
            ArrayList<PseudoTreeNode> temp = new ArrayList<PseudoTreeNode>();
            return temp.iterator();
        }
        else {
            return _childs.keySet().iterator();
        }
    }

    public Iterator<PseudoTreeNode> get_pseudo_parents() {
        if (_pseudo_parents == null) {
            ArrayList<PseudoTreeNode> temp = new ArrayList<PseudoTreeNode>();
            return temp.iterator();
        }
        else {
            return _pseudo_parents.iterator();
        }
    }

    public void setParent(PseudoTreeNode parent, BinaryConstraint bc) {
        if (_parent == null) {
            _parent = parent;
            _parent.addChild(this, bc);
        }
    }
    public void addChild(PseudoTreeNode c, BinaryConstraint bc) {
        if (_childs == null) {
            _childs = new TaggedDictionary<PseudoTreeNode, BinaryConstraint>(this);
        }

        if (!_childs.containsKey(c)) {
            _childs.put(c, bc);
            c.setParent(this, bc);
        }
    }
    public void addPseudo_parent(PseudoTreeNode pp, BinaryConstraint v) {
        if (_pseudo_parents == null) {
            _pseudo_parents = new ArrayList();
        }

        if (!_pseudo_parents.contains(pp)) {
            _pseudo_parents.add(pp);
            pp.addPseudo_child(this, v);
        }
    }
    public void addPseudo_child(PseudoTreeNode pc, BinaryConstraint v) {
        if (_pseudo_childs == null) {
            _pseudo_childs = new TaggedDictionary<PseudoTreeNode, BinaryConstraint>(this);
        }

        if (!_pseudo_childs.containsKey(pc)) {
            _pseudo_childs.put(pc, v);
            pc.addPseudo_parent(this, v);
        }
    }

    public int evaluate_util_weight() {
        Object[] res = _evaluate_util_weight();
        return ((Integer)res[0]).intValue();
    }

    //used only to calculate total weight of the current node, based on how many
    //variables will be evaluated during the util computation
    private Object[] _evaluate_util_weight() {
        ArrayList<Variable> added = new ArrayList<Variable>();
        _util_weight = 0;
        int weight_to_pass = 0;

        if (_childs == null) {
            _eval_weight(null);
            weight_to_pass = _util_weight;

            if (_pseudo_parents != null) {
                added.addAll(_pseudo_parents);
            }
        }
        else {
            //the current node is not a leaf, so send the start signal
            //to all childs
            Iterator it = _childs.keySet().iterator();
            while (it.hasNext()) {
                Object[] partial_res = ((PseudoTreeNode)it.next())._evaluate_util_weight();

                //collects the weight received from childs
                weight_to_pass = weight_to_pass + ((Integer)partial_res[0]).intValue();
                ArrayList<Variable> nodes = (ArrayList<Variable>)partial_res[1];
                
                Iterator<Variable> it_nodes = nodes.iterator();
                while (it_nodes.hasNext()) {
                    PseudoTreeNode current = (PseudoTreeNode)it_nodes.next();
                    if (!current.equals(this) && !added.contains(current)) {
                        added.add(current);
                    }
                }
            }

            //add pseudoparents
            if (_pseudo_parents != null) {
                added.addAll(_pseudo_parents);
            }

            //now evals its weight with the added variables calculated before
            _eval_weight(added);
            weight_to_pass = weight_to_pass + _util_weight;
        }
        
        Object[] res = new Object[2];
        res[0] = new Integer(weight_to_pass);
        res[1] = added;
                
        return res;
    }

    private void _eval_weight(ArrayList<Variable> added) {
        if (_parent != null) {  //the root doesn't evaluate any util, it only selects the max value and begins
            int temp = 1;      //the value propagation phase

            //calculate the weight between its domain size, its parent domain size
            //and its pseudo parents domain size
            if (_pseudo_parents != null) {
                Iterator pp = _pseudo_parents.iterator();
                while (pp.hasNext()) {
                    PseudoTreeNode n = (PseudoTreeNode)pp.next();
                    temp = temp * n.getDomain().size();
                }
            }


            temp = temp * _parent.getDomain().size();


            //add the weight of the added variables
            if (added != null) {
                for (int i = 0; i < added.size(); i++) {
                    if (!added.get(i).equals(_parent) && (_pseudo_parents == null || !_pseudo_parents.contains((Variable)added.get(i)))) {
                        temp = temp * added.get(i).getDomain().size();
                    }
                }
            }

            temp = temp * getDomain().size();

            _util_weight = _util_weight + temp;
        }
    }


    public void start() {
        //the process can start only from the root node
        if (_parent == null) {
            if (_childs == null) {
                //this node isn't connected to any other node, so this node
                //will have a different computation
                _process_isolated_node();
            }
            else {
                _nthreads = 0;
                _nthreadcalls = 1;

                new Thread(this).start();

                //old version without threads:
                //_start_signal();
            }
        }
    }

    public void run() {
        _nthreadcalls--;
        _nthreads++;

        if (_phase == Phase.StartPropagation) {
            _clear_stored_values();

            if (_handler.usePreciseTiming()) {
                _use_precise_timing = true;
            }
            else {
                _use_precise_timing = false;
            }

            _start_signal();
            _nthreads--;
        }
        else if (_phase == Phase.ValuePropagation) {
            _send_value();
            _nthreads--;

            if (_nthreads == 0 && _nthreadcalls == 0) {
                _handler.end();
            }
        }
    }

    private void _start_signal() {
        if (_childs == null) {
            //the procedure can start only if the node is a leaf
            _send_util();
        }
        else {  //_childs != null
            //the current node is not a leaf, so send the start signal
            //to all childs
            Iterator it = _childs.keySet().iterator();
            while (it.hasNext()) {
                _nthreadcalls++;
                new Thread((PseudoTreeNode)it.next()).start();
                //old version without threads:
                //((PseudoTreeNode)it.next())._start_signal();
            }
        }
    }

    public BinaryConstraint getConstraint(PseudoTreeNode n) {
        if (_childs == null && _pseudo_childs == null) {
            return null;
        }
        else if (_childs.containsKey(n)) {
            return (BinaryConstraint)_childs.get(n);
        }
        else if (_pseudo_childs.containsKey(n)) {
            return (BinaryConstraint)_pseudo_childs.get(n);
        }
        else {
            return null;
        }
    }

    private void _send_util() {
        //start propagation phase terminated, util phase begins for the current node
        _phase = Phase.UtilPropagation;

        //first of all collects the relations that must be added in order
        //to calculate the utility for a particular assignment of the
        //variables involved at this node
        ArrayList<Constraint> relations = new ArrayList<Constraint>();

        //then add all the constraints between eventual pseudoparents
        if (_pseudo_parents != null) {
            Iterator pp = _pseudo_parents.iterator();
            while (pp.hasNext()) {
                PseudoTreeNode n = (PseudoTreeNode)pp.next();
                relations.add(n.getConstraint(this));
            }
        }

        //there is at least the relation between the parent and the current node
        relations.add(_parent.getConstraint(this));

        //add all the utils received as constraints
        Iterator<TaggedDictionary<Value, Object>> util_it = _util_received.values().iterator();
        ArrayList<PseudoTreeNode> additional_variables = new ArrayList<PseudoTreeNode>();
        while (util_it.hasNext()) {
            TaggedDictionary<Value, Object> current_util = util_it.next();
            NAryConstraint cons = new NAryConstraint(current_util);
            relations.add(cons);

            //add all the nodes of this util
            ArrayList<Variable> nodes = cons.getVariables();

            Iterator<Variable> it = nodes.iterator();
            while (it.hasNext()) {
                PseudoTreeNode current = (PseudoTreeNode)it.next();
                if (!current.equals(this) && !additional_variables.contains(current)) {
                    additional_variables.add(current);
                }
            }
        }

        //collects all the variables involved in the evaluation of the utility
        //it collects also the domains relative to these variables
        //note: the current variable must be the last added, in order to
        //properly collect the argmax and util value
        ArrayList<PseudoTreeNode> variables = new ArrayList<PseudoTreeNode>();
        //note: the current node and its parent surely will be added

        //add all pseudo parents to the variables
        if (_pseudo_parents != null) {
            Iterator pp = _pseudo_parents.iterator();
            while (pp.hasNext()) {
                PseudoTreeNode n = (PseudoTreeNode)pp.next();
                variables.add(n);
            }
        }

        variables.add(_parent);

        //add the variables that are present in util_received but not in the current
        //variables list. This is necessary because childrens can have more than 1 dimension
        //(because of pseudo parents) and so it will be necessary to propagate these
        //additional dimensions until they collapse in the pseudo_parent variable
        //(worst case in the root node)
        Iterator<PseudoTreeNode> avi = additional_variables.iterator();
        while (avi.hasNext()) {
            PseudoTreeNode current = avi.next();
            if (!variables.contains(current)) {
                variables.add(current);
            }
        }

        if (!variables.contains(this)) {
           variables.add(this);
        }

        //get the iterator that creates the configurations of all possible
        //values depending on the variables domains
        ValuesIterator vi = new ValuesIterator(variables);
        Value current_argmax = null;
        Value current_maxvalue = null;
        Value last_domain_value = vi.getLastValue(this);

        //creates the argmax and the util to send
        //the dimension of these 2 objects depends on how many variables
        //there are at the current node (|variables| - 1)
        //surely there will be at least 1 dimension (1 for the parent of the node)
        PseudoTreeNode first = variables.iterator().next();
        _argmax = new TaggedDictionary<Value, Object>(first);
        TaggedDictionary<Value, Object> util = new TaggedDictionary<Value, Object>(first);
        TaggedDictionary<PseudoTreeNode, Value> current_configuration = new TaggedDictionary<PseudoTreeNode, Value>();

        _handler.evaluating_util(this, variables);

        while(vi.hasNext()) {
            if (!_handler.isProcessing()) {
                //the handler has cancelled the process
                return;
            }

            if (_use_precise_timing) {
                _handler.incrementProgress(1);
            }

            //evaluates the current configuration depending on the values
            //of variables in current_configuration
            current_configuration = vi.next();
            
            //now tests if the current_maxvalue and the current_argmax have to be updated
            Value sum = null;
            for (int j = 0; j < relations.size(); j++) {
                Constraint r = relations.get(j);
                ArrayList<Variable> nodes = r.getVariables();
                
                Value v = null;
                
                //the current relation r is a binary relation between two nodes
                if (r instanceof BinaryConstraint) {
                    Couple c1 = new Couple((PseudoTreeNode)nodes.get(0), current_configuration.get(nodes.get(0)));
                    Couple c2 = new Couple((PseudoTreeNode)nodes.get(1), current_configuration.get(nodes.get(1)));
                    Couple[] c = new Couple[2];
                    c[0] = c1;
                    c[1] = c2;

                    v = r.getConstraint(c);
                    c = null;
                }
                else {
                    //it's a N-ary relation coming from a util received
                    Couple[] c = new Couple[nodes.size()];
                    for (int i = 0; i < nodes.size(); i++) {
                        PseudoTreeNode n = (PseudoTreeNode)nodes.get(i);
                        Couple current_couple = new Couple(n, current_configuration.get(n));
                        c[i] = current_couple;
                    }

                    v = r.getConstraint(c);
                    c = null;
                }

                if (sum == null) {
                    sum = v;
                }
                else {
                    sum = sum.sum(v);
                }
            }
            if (current_maxvalue == null || current_maxvalue.compareTo(sum) < 0) {
                //the sum just calculated is better
                current_maxvalue = sum;
                current_argmax = current_configuration.get(this);
            }

            //if the current configuration has the value for this node
            //reset to the first domain value, then store the current argmax
            // and current max value
            if (last_domain_value.equals(current_configuration.get(this))) {
                _store_argmax_util(util, current_maxvalue, current_argmax, current_configuration, variables);
                current_maxvalue = null;
                current_argmax = null;
            }
        }
        
        if (!_use_precise_timing) {
            _handler.incrementProgress(_util_weight);
        }

        _handler.util_sent(this, _parent);

        //at the end, send util to the parent
        _parent._receive_util(util, this);
    }

    private void _store_argmax_util(TaggedDictionary<Value, Object> util, Value current_maxvalue, Value current_argmax, TaggedDictionary<PseudoTreeNode, Value> current_configuration, ArrayList<PseudoTreeNode> variables) {
        if (current_argmax != null && current_maxvalue != null) {
            //stores the current argmax in _argmax and current max value
            //in util
            Iterator<PseudoTreeNode> keys = current_configuration.keySet().iterator();
            TaggedDictionary<Value, Object> dimension_argmax = _argmax;
            TaggedDictionary<Value, Object> dimension_util = util;
            PseudoTreeNode k;
            PseudoTreeNode k_next;
            k = keys.next();
            while (keys.hasNext()) {
                k_next = keys.next();
                Value v = current_configuration.get(k);
                if (!dimension_util.containsKey(v)) {
                    if (k_next.equals(this)) {
                        dimension_util.put(v, current_maxvalue);
                    }
                    else {
                        TaggedDictionary<Value, Object> newDimension = new TaggedDictionary<Value, Object>(k_next);

                        dimension_util.put(v, newDimension);
                        dimension_util = newDimension;
                    }
                }
                else {
                    TaggedDictionary<Value, Object> newDimension = (TaggedDictionary<Value, Object>)dimension_util.get(v);

                    dimension_util.put(v, newDimension);
                    dimension_util = newDimension;
                }

                if (!dimension_argmax.containsKey(v)) {
                    if (k_next.equals(this)) {
                        dimension_argmax.put(v, current_argmax);
                    }
                    else {
                        TaggedDictionary<Value, Object> newDimension = new TaggedDictionary<Value, Object>(k_next);

                        dimension_argmax.put(v, newDimension);
                        dimension_argmax = newDimension;
                    }
                }
                else {
                    TaggedDictionary<Value, Object> newDimension = (TaggedDictionary<Value, Object>)dimension_argmax.get(v);

                    dimension_argmax.put(v, newDimension);
                    dimension_argmax = newDimension;
                }

                k = k_next;
            }
        }
    }

    private void _receive_util(TaggedDictionary<Value, Object> u, PseudoTreeNode from) {
        _util_received.put(from, u);

        //if all utils received
        if (_util_received.size() == _childs.size()) {
            //if it is the root node
            if (_parent == null) {
                _send_value();
            }
            else {
                _send_util();
            }
        }
    }

    private void _send_value() {
        if (!_handler.isProcessing()) {
            //the handler has cancelled the process
            return;
        }

        if (_received_solution.isEmpty()) {
            //only the root can have the solution empty, because it is the agent
            //who starts the value propagation

            //select the best value from all the utils received from childs
            Value current_maxvalue = null;
            Value current_argmax = null;

            Iterator<TaggedDictionary<Value, Object>> ui = _util_received.values().iterator();
            while (ui.hasNext()) {
                TaggedDictionary<Value, Object> current_util = ui.next();
                Iterator<Value> vi = current_util.keySet().iterator();

                while(vi.hasNext()) {
                    //for each value of the root, evaluates the weight
                    Value k = vi.next();

                    //the util received at the root has obligatorily 1 dimension
                    //(because the root doesn't have parents or pseudo_parents
                    //and all the eventual additional dimensions from childs
                    //collapse in one dimension)
                    Value v = (Value)current_util.get(k);

                    if (current_maxvalue == null || current_maxvalue.compareTo(v) < 0) {
                        current_maxvalue = v;
                        current_argmax = k;
                    }
                }
            }

            //at the end, the partial solution is: root<-current_argmax with
            //weight current_maxvalue
            //send this result to all childs
            TaggedDictionary<PseudoTreeNode, Value> sol = new TaggedDictionary<PseudoTreeNode, Value>(this);
            sol.put(this, current_argmax);

            //send this partial solution also to the handler
            _handler.receive_solution(this, current_argmax);

            //send also the weight to the handler
            _handler.receive_solution_weight(current_maxvalue);

            if (_childs != null) {
                Iterator it = _childs.keySet().iterator();
                while (it.hasNext()) {
                    PseudoTreeNode next = (PseudoTreeNode)it.next();
                    _handler.value_sent(this, next);

                    next._receive_value(sol);
                }
            }
        }
        else {
            //uses the solution received to inspect the _argmax calculated
            //during the util phase in order to find the correct value for the
            //current node
            TaggedDictionary<Value, Object> dimension = _argmax;
            Object new_dimension = null;
            boolean end = false;

            while (!end) {
                PseudoTreeNode n = (PseudoTreeNode)dimension.getTag();
                Value v = _received_solution.get(n);
                new_dimension = dimension.get(v);
                if (new_dimension instanceof Value) {
                    //the solution for the current node is reached
                    _received_solution.put(this, (Value)new_dimension);

                    end = true;
                }
                else {
                    dimension = (TaggedDictionary<Value, Object>)new_dimension;
                }
            }

            //send the solution to the handler
            _handler.receive_solution(this, (Value)new_dimension);

            //-------------------------------------------------------------------
            //addition for the soft graph coloring: info about broken constraints
            //works correctly only with nogood = -1
            Couple c1 = new Couple(this, (Value)new_dimension);
            
            ArrayList<PseudoTreeNode> to_test = new ArrayList<PseudoTreeNode>();
            
            //add the parent
            to_test.add(_parent);
            //add all the eventual pseudo parents
            if (_pseudo_parents != null) {
                Iterator<PseudoTreeNode> pp = _pseudo_parents.iterator();
                while (pp.hasNext()) {
                    to_test.add(pp.next());
                }
            }
            
            Iterator<PseudoTreeNode> it_test = to_test.iterator();
            while (it_test.hasNext()) {
                PseudoTreeNode current = it_test.next();
                
                Couple ci = new Couple(current, _received_solution.get(current));
                    
                Couple[] arr = new Couple[2];
                arr[0] = c1;
                arr[1] = ci;
                if (current.getConstraint(this).getConstraint(arr).equals(new IntegerValue(-1))) {
                    _handler.broken_constraint(current, this);
                }
            }
            //---------------------------------------------------------------------

            //if the current node is a leaf, send the solution to the handler
            //otherwise send it to all childs
            if (_childs != null) {
                Iterator it = _childs.keySet().iterator();
                while (it.hasNext()) {
                    PseudoTreeNode next = (PseudoTreeNode)it.next();
                    _handler.value_sent(this, next);

                    next._receive_value(_received_solution);
                }
            }
        }

        //after the send value phase, the phase is reset for the future computations
        _phase = Phase.StartPropagation;
    }

    private void _receive_value(TaggedDictionary<PseudoTreeNode, Value> sol) {
        //util propagation phase has terminated, value phase begins
        _phase = Phase.ValuePropagation;

        _received_solution = sol;

        _nthreadcalls++;
        new Thread(this).start();

        //old version without threads
        //_send_value();
    }

    private void _process_isolated_node() {
        //the solution of the current node is simply the first value of the
        //domain, because the node is not connected, so any value will be ok
        Value v = getDomain().get(0);
        
        _handler.receive_solution(this, v);
        _handler.value_sent(this, null);
    }


    @Override
    public String toString() {
        return "variable " + getId().toString() + ", domain: " + getDomain().toString();
    }

    public String shortToString() {
        return "x" + getId().toString();
    }
}
