/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package dpop;

import java.util.*;
import java.lang.*;
import java.io.*;

/**
 *
 * @author Andrea
 */
public class TaggedDictionary<K, V> implements Map<K, V> {
    private Object _tag;
    private MyEntrySet<K, V> _es;

    private final class MyEntrySet<K, V> extends AbstractSet<Map.Entry<K,V>> {
        private ArrayList<K> _keys;
        private ArrayList<V> _values;
    
        MyEntrySet() {
            _keys = new ArrayList<K>();
            _values = new ArrayList<V>();
        }
        
        @Override
        public Iterator<Map.Entry<K, V>> iterator() {
            return new MyIterator();
        }

        @Override
        public int size() {
            return _keys.size();
        }

    }
    private final class MyIterator<K, V> implements Iterator<Map.Entry<K, V>> {
        private int _pos;
        private int _last_returned;
        
        MyIterator() {
            _pos = 0;
            _last_returned = -1;
        }
        
        public boolean hasNext() {
            return _pos != _es._keys.size();
        }

        public Map.Entry<K, V> next() {
            V currentValue = (V)_es._values.get(_pos);
            if (currentValue == null) {
                return null;
            }
            else {
                K currentKey = (K)_es._keys.get(_pos);
                _last_returned = _pos++;
                return new MyEntry(currentKey, currentValue);
            }
        }

        public void remove() {
            if (_last_returned == -1) {
		throw new IllegalStateException();
            }
            
            _es._keys.remove(_last_returned);
            _es._values.remove(_last_returned);
            
            if (_last_returned < _pos) {
		_pos--;
            }
            _last_returned = -1;
        }
        
    }
    private final class MyEntry<K, V> implements Map.Entry<K,V> {
        private K _key;
        private V _value;

        MyEntry(K key, V value) {
            _key = key;
            _value = value;
        }

        public K getKey() {
            return _key;
        }

        public V getValue() {
            return _value;
        }

        public V setValue(V value) {
            V old = _value;
            _value = value;
            return old;
        }

    }
    private final class MyKeySet<K> extends AbstractSet<K> {
        MyKeySet() {
        }

        public Iterator<K> iterator() {
            return new MyKeyIterator();
        }
        public int size() {
            return _es.size();
        }

        public boolean contains(Object o) {
            return _es._keys.contains(o);
        }

        public boolean remove(Object o) {
            int i = _es._keys.indexOf(o);
            if (i >= 0) {
                V v = _es._values.get(i);
                _es._values.remove(i);
                _es._keys.remove(i);
                return true;
            }
            else {
                return false;
            }
        }

        @Override
        public void clear() {
            _es.clear();
        }
    }
    private final class MyKeyIterator<K> implements Iterator<K> {
        private int _pos;
        private int _last_returned;

        MyKeyIterator() {
            _pos = 0;
            _last_returned = -1;
        }

        public boolean hasNext() {
            return _pos != _es._keys.size();
        }

        public K next() {
            K currentKey = (K)_es._keys.get(_pos);
            _last_returned = _pos++;
            return currentKey;
        }

        public void remove() {
            if (_last_returned == -1) {
		throw new IllegalStateException();
            }

            _es._keys.remove(_last_returned);

            if (_last_returned < _pos) {
		_pos--;
            }
            _last_returned = -1;
        }
    }

    TaggedDictionary() {
        _tag = null;
        _es = new MyEntrySet();
    }

    TaggedDictionary(Object tag) {
        _tag = tag;
        _es = new MyEntrySet();
    }

    public Object getTag() {
        return _tag;
    }

    public void clear() {
        _es.clear();
    }

    public boolean containsKey(Object key) {
        return _es._keys.contains((K)key);
    }

    public boolean containsValue(Object value) {
        return _es._values.contains((V)value);
    }

    public Set<Map.Entry<K, V>> entrySet() {
        return _es;
    }

    public V get(Object key) {
        int i = _es._keys.indexOf(key);
        return (i == -1) ? null : _es._values.get(i);
    }

    public boolean isEmpty() {
        return _es.isEmpty();
    }

    public Set<K> keySet() {
        return new MyKeySet();
    }

    public V put(K key, V value) {
        int i = _es._keys.indexOf(key);
        if (i >= 0) {
            V prev = _es._values.get(i);
            _es._values.set(i, value);
            return prev;
        }
        else {
            _es._keys.add(key);
            _es._values.add(value);
            return null;
        }
    }

    public void putAll(Map<? extends K,? extends V> m) {
        Set<? extends Map.Entry<? extends K, ? extends V>> s = m.entrySet();
        Iterator<? extends Map.Entry<? extends K, ? extends V>> i = s.iterator();
        while (i.hasNext()) {
            Map.Entry<? extends K, ? extends V> e = i.next();
            put(e.getKey(), e.getValue());
        }
    }

    public V remove(Object key) {
        MyKeySet<K> ks = new MyKeySet();
        ks.remove(key);
        return null;
    }

    public int size() {
        return _es.size();
    }

    public Collection values() {
        return _es._values;
    }
}
